# -*- coding: utf-8 -*-
from functools import wraps
import re

from django.db import transaction
from django.conf import settings


class MasterSlaveRouter(object):
    read_from_master = False # A global flag to switch read requests to master

    _replication_sets_re = set(re.compile(s[0]) for s in settings.DATABASE_SETS_RE)
    _cache = set()
    def _is_replicated(self, model):
        table_name = model._meta.db_table
        if table_name in self._cache:
            return True
        for regex in self._replication_sets_re:
            if regex.search(table_name):
                self._cache.add(table_name)
                return True
        return False

    def db_for_read(self, model, **hints):
        if self.read_from_master:
            return self.db_for_write(model, **hints)
        return 'default'

    def db_for_write(self, model, **hints):
        return settings.DEFAULT_MASTER if self._is_replicated(model) else 'default'

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_syncdb(self, db, model):
        return True


def read_from_master(func):
    """
    Decorator which switches any read db requests done in function to master.
    Should be used with handy.db.MasterSlaveRouter.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        previous_state = MasterSlaveRouter.read_from_master
        MasterSlaveRouter.read_from_master = True
        try:
            return func(*args, **kwargs)
        except:
            raise
        finally:
            MasterSlaveRouter.read_from_master = previous_state
    return wrapper


def commit_on_success_on(*databases):
    databases = set(databases)
    def decorator(func):
        for db in databases:
            func = transaction.commit_on_success(db)(func)
        return wraps(func)(func)
    return decorator



# -*- coding: utf-8 -*-
import re

class StripWhitespace(object):
    """
    Strips leading and trailing whitespace from response content.
    """
    NEWLINE_RE = re.compile(r'>\s*\n\s*<')
    SPACES_RE = re.compile(r'>\s{2,}<')

    def process_response(self, request, response):
        if response['Content-Type'].startswith('text/html'):
            content = response.content
            content = self.NEWLINE_RE.sub('>\n<', content)
            content = self.SPACES_RE.sub('> <', content)
            response.content = content
        return response



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
from functools import wraps


class SkipMemoization(Exception):
    pass


def memoize(func):
    cache = {}

    @wraps(func)
    def wrapper(*args):
        if args not in cache:
            try:
                cache[args] = func(*args)
            except SkipMemoization, e:
                return e.args[0] if e.args else None

        return cache[args]
    return wrapper
memoize.skip = SkipMemoization


def cache(timeout):
    if isinstance(timeout, int):
        timeout = timedelta(seconds=timeout)

    def decorator(func):
        cache = {}

        @wraps(func)
        def wrapper(*args):
            if args in cache:
                result, timestamp = cache[args]
                if datetime.now() - timestamp < timeout:
                    return result
                else:
                    del cache[args]

            result = func(*args)
            cache[args] = result, datetime.now()
            return result

        def invalidate(*args):
            cache.pop(args)
        wrapper.invalidate = invalidate

        return wrapper
    return decorator



# -*- coding: utf-8 -*-
from datetime import datetime
from django.utils.importlib import import_module


from django.db.models.fields.related import ForeignKey
ForeignKey.__dump__ = lambda self: 'ForeignKey(%s.%s -> %s.%s)' % (self.model, self.attname, self.rel.to, self.rel.field_name)

from django.db.models.sql.where import Constraint
Constraint.__dump__ = lambda self: 'Constraint(%s.%s)' % (self.alias, self.col)


def obj_dump(obj, indent=0, seen=None):
    if seen is None:
        seen = set()
    result = []
    if obj is None or isinstance(obj, (str, int, long, float, bool, dict, set, datetime)) or callable(obj):
        result.append(repr(obj))
    elif isinstance(obj, unicode):
        result.append("u'" + obj.encode('utf8') + "'")
    elif isinstance(obj, (list, tuple)) and not obj:
        result.append(repr(obj))
    elif hasattr(obj, '__dump__'):
        result.append(obj.__dump__())
    elif id(obj) in seen:
        result.append('<SEEN_OBJECT>')
    else:
        seen.add(id(obj))
        if isinstance(obj, (list, tuple)):
            result.append('[\n' if isinstance(obj, list) else '(\n')
            for item in obj:
                result.append('%s,' % obj_dump(item, indent=indent+4, seen=seen))
            result.append(']' if isinstance(obj, list) else ')')
        else:
            result.append(repr(obj) + '\n')
            for attr, value in obj.__dict__.iteritems():
                result.append('%s: %s,' % (attr, obj_dump(value, indent=indent+4, seen=seen)))

    result = result[0] + "\n".join(' ' * indent + i for i in result[1:])
    return result


def get_module_attr(path):
    """
    Возвращает атрибут модуля, переданный в виде строки 'module.other.AttrName'.
    Удобно для получения класса по строке импорта.
    При отсутствии атрибута возвращает None.
    """
    i = path.rfind('.')
    module, attr = path[:i], path[i+1:]
    try:
        mod = import_module(module)
        return getattr(mod, attr, None)
    except ImportError, e:
        return None


def cached_property(func):
    @property
    @wraps(func)
    def wrapper(self):
        attname = '_' + func.__name__
        if not hasattr(self, attname):
            setattr(self, attname, func(self))
        return getattr(self, attname)
    return wrapper



# -*- coding: utf-8 -*-
import re, datetime
from functools import wraps
import pytz

from .decorators import render_to_json


class AjaxException(Exception):
    def __init__(self, message='', **kwargs):
        defaults = {'success': False, 'error': message, 'data': {}}
        self.params = dict(defaults, **kwargs)

class Ajax(object):
    error = AjaxException

    def __call__(self, func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            try:
                data = func(request, *args, **kwargs)
                return {'success': True, 'data': data}
            except self.error, e:
                return e.params
        return render_to_json(default=encode_object)(wrapper)

    def catch(self, *exceptions):
        assert exceptions, 'at least one exception type is required'
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except self.error:
                    raise
                except exceptions, e:
                    error_name = e.__class__.__name__
                    error_text = e.args[0] if e.args else ''
                    return {'success': False,
                            'error': camel_to_underscores(error_name),
                            'error_text': error_text}
            return wrapper
        return decorator

    def login_required(self, func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            if request.user.is_anonymous():
                raise self.error('login_required')
            else:
                return func(request, *args, **kwargs)
        return wrapper

ajax = Ajax()


def camel_to_underscores(name):
    words = re.findall(r'[A-Z][a-z0-9]+', name)
    return '_'.join(w.lower() for w in words)


### Encoding object to json
from pytz.tzinfo import DstTzInfo
DstTzInfo.__json__ = lambda self: self.zone

UTC = pytz.timezone('UTC')
EPOCH = datetime.datetime(1970,1,1,0,0,0, tzinfo=UTC)

# TODO: add support for querysets and models
def encode_object(obj):
    if hasattr(obj, '__json__'):
        return obj.__json__()
    elif isinstance(obj, datetime.datetime):
        delta = timezone.to_default(obj) - EPOCH
        return {'_type': 'date', 'ms': int(delta.total_seconds() * 1000)}
    elif isinstance(obj, datetime.date):
        return datetime.datetime(*obj.timetuple()[:6])
    else:
        raise TypeError('%s is not JSON serializable' % obj.__class__.__name__)



# -*- coding: utf-8 -*-
import logging, logging.handlers

from django.conf import settings


def get_logger(name, level=logging.INFO, format='[%(asctime)s] %(message)s', handler=None, filename=None):
    new_logger = logging.getLogger(name)
    new_logger.setLevel(level)

    if not handler:
        filename = filename or '%s/logs/%s.log' % (settings.HOME_DIR, name)
        handler = logging.FileHandler(filename)

    handler.setFormatter(logging.Formatter(format))
    new_logger.addHandler(handler)

    return new_logger


if hasattr(settings, 'LOG_FILENAME') and not logger:
    handler = logging.handlers.TimedRotatingFileHandler(settings.LOG_FILENAME, when = 'midnight')
    logger = get_logger('default', handler=handler)



from setuptools import setup

setup(
    name='handy',
    version='0.1.2',
    author='Alexander Schepanovski',
    author_email='suor.web@gmail.com',

    description='A collection of tools to make your django life easier.',
    long_description=open('README.rst').read(),
    url='http://github.com/Suor/handy',
    license='BSD',

    packages=['handy'],
    install_requires=[
        'django>=1.2',
    ],

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',

        'Framework :: Django',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
